@@ -37,6 +37,8 @@
 #define GESTURE_LAG_MODE              8
 #define GESTURE_PREDRAG_MODE         16
 #define GESTURE_DRAG_MODE            32
+#define GESTURE_CANCEL_MODE          64
+#define GESTURE_MULTITOUCH_MODE     128
 
 #define WCM_SCROLL_UP                 5	/* vertical up */
 #define WCM_SCROLL_DOWN               4	/* vertical down */
@@ -101,43 +103,84 @@ static void getStateHistory(WacomCommonPtr common, WacomDeviceState states[], in
  * the multitouch API available in XI2.2.
  *
  * @param[in] priv
- * @param[in] contact_id  ID of the contact to send event for
+ * @param[in] channel    Channel to send a touch event for
+ * @param[in] no_update  If 'true', TouchUpdate events will not be created.
+ * This should be used when entering multitouch mode to ensure TouchBegin
+ * events are sent for already-in-prox contacts.
  */
 static void
-wcmSendTouchEvent(WacomDevicePtr priv, int contact_id)
+wcmSendTouchEvent(WacomDevicePtr priv, WacomChannelPtr channel, Bool no_update)
 {
 #if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 16
-	WacomChannelPtr channel = getContactNumber(priv->common, contact_id);
+	ValuatorMask *mask = priv->common->touch_mask;
+	WacomDeviceState state = channel->valid.state;
+	WacomDeviceState oldstate = channel->valid.states[1];
+	int type = -1;
 
-	if (channel) {
-		WacomDeviceState state  = channel->valid.state;
-		ValuatorMask *mask = priv->common->touch_mask;
-		WacomDeviceState oldstate = channel->valid.states[1];
-		int type = -1;
+	wcmRotateAndScaleCoordinates (priv->pInfo, &state.x, &state.y);
 
-		wcmRotateAndScaleCoordinates (priv->pInfo, &state.x, &state.y);
+	valuator_mask_set(mask, 0, state.x);
+	valuator_mask_set(mask, 1, state.y);
 
-		valuator_mask_set(mask, 0, state.x);
-		valuator_mask_set(mask, 1, state.y);
-
-		if (!state.proximity) {
-			DBG(6, priv->common, "This is a touch end event\n");
-			type = XI_TouchEnd;
-		}
-		else if (!oldstate.proximity) {
-			DBG(6, priv->common, "This is a touch begin event\n");
-			type = XI_TouchBegin;
-		}
-		else {
-			DBG(6, priv->common, "This is a touch update event\n");
-			type = XI_TouchUpdate;
-		}
-
-		xf86PostTouchEvent(priv->pInfo->dev, contact_id, type, 0, mask);
+	if (!state.proximity) {
+		DBG(6, priv->common, "This is a touch end event\n");
+		type = XI_TouchEnd;
 	}
+	else if (!oldstate.proximity || no_update) {
+		DBG(6, priv->common, "This is a touch begin event\n");
+		type = XI_TouchBegin;
+	}
+	else {
+		DBG(6, priv->common, "This is a touch update event\n");
+		type = XI_TouchUpdate;
+	}
+
+	xf86PostTouchEvent(priv->pInfo->dev, state.serial_num - 1, type, 0, mask);
 #endif
 }
 
+/**
+ * Send multitouch events. If entering multitouch mode (indicated by
+ * GESTURE_LAG_MODE), then touch events are sent for all in-prox
+ * contacts. Otherwise, only the specified contact has a touch event
+ * generated.
+ *
+ * @param[in] priv
+ * @param[in] contact_id  ID of the contact to send event for (at minimum)
+ */
+static void
+wcmFingerMultitouch(WacomDevicePtr priv, int contact_id) {
+	Bool lag_mode = priv->common->wcmGestureMode == GESTURE_LAG_MODE;
+	Bool prox = FALSE;
+	int i;
+
+	if (lag_mode && TabletHasFeature(priv->common, WCM_LCD)) {
+		/* wcmSingleFingerPress triggers a button press as
+		 * soon as a single finger appears. ensure we release
+		 * that button before getting too far along
+		 */
+		wcmSendButtonClick(priv, 1, 0);
+	}
+
+	for (i = 0; i < MAX_CHANNELS; i++) {
+		WacomChannelPtr channel = priv->common->wcmChannel+i;
+		WacomDeviceState state  = channel->valid.state;
+		if (state.device_type != TOUCH_ID)
+			continue;
+
+		if (lag_mode || state.serial_num == contact_id + 1) {
+			wcmSendTouchEvent(priv, channel, lag_mode);
+		}
+
+		prox |= state.proximity;
+	}
+
+	if (!prox)
+		priv->common->wcmGestureMode = GESTURE_NONE_MODE;
+	else if (lag_mode)
+		priv->common->wcmGestureMode = GESTURE_MULTITOUCH_MODE;
+}
+
 static double touchDistance(WacomDeviceState ds0, WacomDeviceState ds1)
 {
 	int xDelta = ds0.x - ds1.x;
@@ -198,7 +241,7 @@ static void wcmSendButtonClick(WacomDevicePtr priv, int button, int state)
 	/* We have changed the button state (from down to up) for the device
 	 * so we need to update the record */
 	if (button == 1)
-		priv->oldButtons = 0;
+		priv->oldState.buttons = 0;
 }
 
 /*****************************************************************************
@@ -327,21 +370,30 @@ static void wcmSingleFingerPress(WacomDevicePtr priv)
 	}
 }
 
+
+/**
+ * Cancel any in-progress gesture, returning to GESTURE_NONE_MODE until new
+ * fingers enter proximity.
+ */
+void wcmCancelGesture(InputInfoPtr pInfo)
+{
+	WacomDevicePtr priv = pInfo->private;
+	WacomCommonPtr common = priv->common;
+
+	if (!IsTouch(priv))
+		return;
+
+	if (common->wcmGestureMode == GESTURE_DRAG_MODE)
+		wcmSendButtonClick(priv, 1, 0);
+	common->wcmGestureMode = GESTURE_CANCEL_MODE;
+}
+
 /* parsing gesture mode according to 2FGT data */
 void wcmGestureFilter(WacomDevicePtr priv, int touch_id)
 {
 	WacomCommonPtr common = priv->common;
 	WacomDeviceState ds[2] = {{0}}, dsLast[2] = {{0}};
 
-#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 16
-	/* Send multitouch data to X if appropriate */
-	if (!common->wcmGesture)
-	{
-		wcmSendTouchEvent(priv, touch_id);
-		return;
-	}
-#endif
-
 	getStateHistory(common, ds, ARRAY_SIZE(ds), 0);
 	getStateHistory(common, dsLast, ARRAY_SIZE(dsLast), 1);
 
@@ -355,6 +407,20 @@ void wcmGestureFilter(WacomDevicePtr priv, int touch_id)
 		return;
 	}
 
+	/* Do not process gestures while in CANCEL mode. Only reset back to
+	 * NONE mode once all fingers have left the screen.
+	 */
+	if (common->wcmGestureMode == GESTURE_CANCEL_MODE)
+	{
+		if (ds[0].proximity || ds[1].proximity)
+			return;
+		else
+			common->wcmGestureMode = GESTURE_NONE_MODE;
+	}
+
+	if (common->wcmGestureMode == GESTURE_MULTITOUCH_MODE)
+		goto ret;
+
 	/* When 2 fingers are in proximity, it must always be in one of
 	 * the valid 2 fingers modes: LAG, SCROLL, or ZOOM.
 	 * LAG mode is used while deciding between SCROLL and ZOOM and
@@ -487,6 +553,22 @@ void wcmGestureFilter(WacomDevicePtr priv, int touch_id)
 	}
 ret:
 
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 16
+	/* Send multitouch data to X if appropriate */
+	if (!common->wcmGesture) {
+		if (common->wcmGestureMode == GESTURE_NONE_MODE) {
+			if (TabletHasFeature(common, WCM_LCD))
+				common->wcmGestureMode = GESTURE_MULTITOUCH_MODE;
+			else if (ds[1].proximity)
+				common->wcmGestureMode = GESTURE_LAG_MODE;
+		}
+
+		if (common->wcmGestureMode == GESTURE_LAG_MODE ||
+		    common->wcmGestureMode == GESTURE_MULTITOUCH_MODE)
+			wcmFingerMultitouch(priv, touch_id);
+	}
+#endif
+
 	if ((common->wcmGestureMode == GESTURE_NONE_MODE || common->wcmGestureMode == GESTURE_DRAG_MODE) &&
 	    touch_id == 0)
 	{
