@@ -41,6 +41,7 @@ typedef struct {
 	int nbuttons;                /* total number of buttons */
 	int npadkeys;                /* number of pad keys in the above array */
 	int padkey_code[WCM_MAX_BUTTONS];/* hardware codes for buttons */
+	int lastChannel;
 } wcmUSBData;
 
 static Bool usbDetect(InputInfoPtr);
@@ -302,7 +303,12 @@ static struct WacomModelDesc
 	{ WACOM_VENDOR_ID, 0x100,100000, 100000, &usbTabletPC,   "TabletPC 0x100"	},
 	{ WACOM_VENDOR_ID, 0x101,100000, 100000, &usbTabletPC,   "TabletPC 0x101"	},
 	{ WACOM_VENDOR_ID, 0x10D,100000, 100000, &usbTabletPC,   "TabletPC 0x10D"	},
+	{ WACOM_VENDOR_ID, 0x116,100000, 100000, &usbTabletPC,   "TabletPC 0x116"	},
+	{ WACOM_VENDOR_ID, 0x12C,100000, 100000, &usbTabletPC,   "TabletPC 0x12C"	},
 	{ WACOM_VENDOR_ID, 0x4001,100000, 100000, &usbTabletPC,  "TabletPC 0x4001"	},
+	{ WACOM_VENDOR_ID, 0x4004,100000, 100000, &usbTabletPC,  "TabletPC 0x4004"	},
+	{ WACOM_VENDOR_ID, 0x5000,100000, 100000, &usbTabletPC,  "TabletPC 0x5000"	},
+	{ WACOM_VENDOR_ID, 0x5002,100000, 100000, &usbTabletPC,  "TabletPC 0x5002"	},
 
 	/* IDs from Waltop's driver, available http://www.waltop.com.tw/download.asp?lv=0&id=2.
 	   Accessed 8 Apr 2010, driver release date 2009/08/11, fork of linuxwacom 0.8.4.
@@ -567,6 +573,7 @@ int usbWcmGetRanges(InputInfoPtr pInfo)
 
 	if (!is_touch)
 	{
+		common->wcmMinX = absinfo.minimum;
 		common->wcmMaxX = absinfo.maximum;
 
 #if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,30)
@@ -600,6 +607,7 @@ int usbWcmGetRanges(InputInfoPtr pInfo)
 
 	if (!is_touch)
 	{
+		common->wcmMinY = absinfo.minimum;
 		common->wcmMaxY = absinfo.maximum;
 
 #if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,30)
@@ -837,8 +845,8 @@ static int protocol5Serial(int device_type, unsigned int serial) {
  * Find an appropriate channel to track the specified tool's state in.
  * If the tool is already in proximity, the channel currently being used
  * to store its state will be returned. Otherwise, an arbitrary available
- * channel will be returned. Up to MAX_CHANNEL tools can be tracked
- * concurrently by driver.
+ * channel will be cleaned and returned. Up to MAX_CHANNEL tools can be
+ * tracked concurrently by driver.
  *
  * @param[in] common
  * @param[in] device_type  Type of tool (e.g. STYLUS_ID, TOUCH_ID, PAD_ID)
@@ -869,7 +877,7 @@ static int usbChooseChannel(WacomCommonPtr common, int device_type, unsigned int
 		}
 	}
 
-	/* find an empty channel */
+	/* find and clean an empty channel */
 	if (channel < 0)
 	{
 		for (i=0; i<MAX_CHANNELS; i++)
@@ -880,6 +888,7 @@ static int usbChooseChannel(WacomCommonPtr common, int device_type, unsigned int
 			if (!common->wcmChannel[i].work.proximity)
 			{
 				channel = i;
+				memset(&common->wcmChannel[channel],0, sizeof(WacomChannel));
 				break;
 			}
 		}
@@ -1087,6 +1096,9 @@ static int usbIdToType(int id)
 {
 	int type = STYLUS_ID;
 
+	if (!id)
+		return 0;
+
 	/* The existing tool ids have the following patten: all pucks, except
 	 * one, have the third byte set to zero; all erasers have the fourth
 	 * bit set. The rest are styli.
@@ -1100,56 +1112,32 @@ static int usbIdToType(int id)
 }
 
 /**
- * Find the tool type (STYLUS_ID, etc.) based on the device_id or the
- *  current tool serial number if the device_id is unknown (0).
+ * Find the tool type (STYLUS_ID, etc.) based on the device_id.
  *
  * Protocol 5 devices report different IDs for different styli and pucks,
  * Protocol 4 devices simply report STYLUS_DEVICE_ID, etc.
  *
- * @param ds The current device state received from the kernel.
- * @return The tool type associated with the tool id or the current
- * tool serial number.
+ * @device_id id of the device
+ * @return device type
  */
-static int usbFindDeviceType(const WacomCommonPtr common,
-			  const WacomDeviceState *ds)
+static int usbFindDeviceTypeById(int device_id)
 {
-	WacomToolPtr tool = NULL;
-	int device_type = 0;
-
-	if (!ds->device_id && ds->serial_num)
-	{
-		for (tool = common->wcmTool; tool; tool = tool->next)
-			if (ds->serial_num == tool->serial)
-			{
-				device_type = tool->typeid;
-				break;
-			}
-	}
-
-	if (device_type || !ds->device_id) return device_type;
-
-	switch (ds->device_id)
+	switch (device_id)
 	{
 		case STYLUS_DEVICE_ID:
-			device_type = STYLUS_ID;
-			break;
+			return STYLUS_ID;
 		case ERASER_DEVICE_ID:
-			device_type = ERASER_ID;
-			break;
+			return ERASER_ID;
 		case CURSOR_DEVICE_ID:
-			device_type = CURSOR_ID;
-			break;
+			return CURSOR_ID;
 		case TOUCH_DEVICE_ID:
-			device_type = TOUCH_ID;
-			break;
+			return TOUCH_ID;
 		case PAD_DEVICE_ID:
-			device_type = PAD_ID;
-			break;
+			return PAD_ID;
 		default: /* protocol 5 */
-			device_type = usbIdToType(ds->device_id);
+			return usbIdToType(device_id);
 	}
-
-	return device_type;
+	return 0;
 }
 
 static void usbParseAbsEvent(WacomCommonPtr common,
@@ -1207,15 +1195,13 @@ static void usbParseAbsEvent(WacomCommonPtr common,
 		case ABS_MISC:
 			ds->proximity = (event->value != 0);
 			if (event->value)
-			{
 				ds->device_id = event->value;
-				ds->device_type = usbFindDeviceType(common, ds);
-			}
 			break;
 		default:
 			change = 0;
 	}
 
+	ds->time = (int)GetTimeInMillis();
 	channel->dirty |= change;
 }
 
@@ -1272,6 +1258,7 @@ static void usbParseAbsMTEvent(WacomCommonPtr common, struct input_event *event)
 
 		case ABS_MT_TRACKING_ID:
 			ds->proximity = (event->value != -1);
+			/* set this here as type for this channel doesn't get set in usbDispatchEvent() */
 			ds->device_type = TOUCH_ID;
 			ds->device_id = TOUCH_DEVICE_ID;
 			ds->sample = (int)GetTimeInMillis();
@@ -1293,6 +1280,7 @@ static void usbParseAbsMTEvent(WacomCommonPtr common, struct input_event *event)
 			change = 0;
 	}
 
+	ds->time = (int)GetTimeInMillis();
 	(&common->wcmChannel[private->wcmMTChannel])->dirty |= change;
 }
 
@@ -1315,9 +1303,8 @@ static void usbParseKeyEvent(WacomCommonPtr common,
 		case BTN_TOOL_PENCIL:
 		case BTN_TOOL_BRUSH:
 		case BTN_TOOL_AIRBRUSH:
-			ds->device_type = STYLUS_ID;
 			/* V5 tools use ABS_MISC to report device_id */
-			if (common->wcmProtocolLevel == WCM_PROTOCOL_4)
+			if (common->wcmProtocolLevel != WCM_PROTOCOL_5)
 				ds->device_id = STYLUS_DEVICE_ID;
 			ds->proximity = (event->value != 0);
 			DBG(6, common,
@@ -1326,13 +1313,10 @@ static void usbParseKeyEvent(WacomCommonPtr common,
 			break;
 
 		case BTN_TOOL_RUBBER:
-			ds->device_type = ERASER_ID;
 			/* V5 tools use ABS_MISC to report device_id */
-			if (common->wcmProtocolLevel == WCM_PROTOCOL_4)
+			if (common->wcmProtocolLevel != WCM_PROTOCOL_5)
 				ds->device_id = ERASER_DEVICE_ID;
 			ds->proximity = (event->value != 0);
-			if (ds->proximity)
-				ds->proximity = ERASER_PROX;
 			DBG(6, common,
 			    "USB eraser detected %x (value=%d)\n",
 			    event->code, event->value);
@@ -1343,9 +1327,8 @@ static void usbParseKeyEvent(WacomCommonPtr common,
 			DBG(6, common,
 			    "USB mouse detected %x (value=%d)\n",
 			    event->code, event->value);
-			ds->device_type = CURSOR_ID;
 			/* V5 tools use ABS_MISC to report device_id */
-			if (common->wcmProtocolLevel == WCM_PROTOCOL_4)
+			if (common->wcmProtocolLevel != WCM_PROTOCOL_5)
 				ds->device_id = CURSOR_DEVICE_ID;
 			ds->proximity = (event->value != 0);
 			break;
@@ -1361,7 +1344,6 @@ static void usbParseKeyEvent(WacomCommonPtr common,
 					DBG(6, common,
 					    "USB 1FG Touch detected %x (value=%d)\n",
 					    event->code, event->value);
-					ds->device_type = TOUCH_ID;
 					ds->device_id = TOUCH_DEVICE_ID;
 					ds->proximity = event->value;
 				}
@@ -1375,7 +1357,6 @@ static void usbParseKeyEvent(WacomCommonPtr common,
 				DBG(6, common,
 				    "USB Pad detected %x (value=%d)\n",
 				event->code, event->value);
-				ds->device_type = PAD_ID;
 				ds->device_id = PAD_DEVICE_ID;
 				ds->proximity = (event->value != 0);
 				break;
@@ -1386,7 +1367,6 @@ static void usbParseKeyEvent(WacomCommonPtr common,
 			DBG(6, common,
 			    "USB Touch detected %x (value=%d)\n",
 			    event->code, event->value);
-			ds->device_type = TOUCH_ID;
 			ds->device_id = TOUCH_DEVICE_ID;
 			ds->proximity = event->value;
 			/* time stamp for 2FGT gesture events */
@@ -1399,7 +1379,6 @@ static void usbParseKeyEvent(WacomCommonPtr common,
 			DBG(6, common,
 			    "USB Touch second finger detected %x (value=%d)\n",
 			    event->code, event->value);
-			ds->device_type = TOUCH_ID;
 			ds->device_id = TOUCH_DEVICE_ID;
 			ds->proximity = event->value;
 			/* time stamp for 2GT gesture events */
@@ -1414,6 +1393,7 @@ static void usbParseKeyEvent(WacomCommonPtr common,
 			change = 0;
 	}
 
+	ds->time = (int)GetTimeInMillis();
 	channel->dirty |= change;
 
 	if (change)
@@ -1440,6 +1420,7 @@ static void usbParseKeyEvent(WacomCommonPtr common,
 			change = 0;
 	}
 
+	ds->time = (int)GetTimeInMillis();
 	channel->dirty |= change;
 }
 
@@ -1490,23 +1471,21 @@ static void usbParseBTNEvent(WacomCommonPtr common,
 				change = 0;
 	}
 
+	ds->time = (int)GetTimeInMillis();
 	channel->dirty |= change;
-
-	/* expresskey pressed at startup or missing type */
-	if (!ds->device_type && channel->dirty)
-		ds->device_type = PAD_ID;
 }
 
 /**
- * Translates a tool code from the kernel (e.g. BTN_TOOL_PEN) into the
- * corresponding device type for the driver (e.g. STYLUS_ID).
+ * Translates an event code from the kernel (e.g. type: EV_ABS code: ABS_MISC value: STYLUS_DEVICE_ID)
+ * into the corresponding device type for the driver (e.g. STYLUS_ID).
  *
  * @param[in] common
  * @param[in] type      Linux input tool type (e.g. EV_KEY)
  * @param[in] code      Linux input tool code (e.g. BTN_STYLUS_PEN)
+ * @param[in] value     Linux input tool value (e.g. STYLUS_DEVICE_ID)
  * @return              Wacom device ID (e.g. STYLUS_ID) or 0 if no match.
  */
-static int toolTypeToDeviceType(WacomCommonPtr common, int type, int code)
+static int deviceTypeFromEvent(WacomCommonPtr common, int type, int code, int value)
 {
 	wcmUSBData* private = common->private;
 
@@ -1524,13 +1503,27 @@ static int toolTypeToDeviceType(WacomCommonPtr common, int type, int code)
 
 			case BTN_TOOL_FINGER:
 				if ((common->wcmProtocolLevel != WCM_PROTOCOL_GENERIC)
-				    && !private->wcmUseMT)
+				    && !private->wcmUseMT)  /* this isn't in usbParseKeyEvent() */
 					return PAD_ID;
 				else
 					return TOUCH_ID;
 
 			case BTN_TOOL_RUBBER:
 				return ERASER_ID;
+
+			case BTN_TOOL_DOUBLETAP:
+			case BTN_TOOL_TRIPLETAP:
+				return TOUCH_ID;
+
+			case BTN_TOUCH:
+				if (common->wcmProtocolLevel == WCM_PROTOCOL_GENERIC &&
+				    /* 1FG USB touchscreen */
+				    !TabletHasFeature(common, WCM_PEN) &&
+				    TabletHasFeature(common, WCM_1FGT) &&
+				    TabletHasFeature(common, WCM_LCD))
+					return TOUCH_ID;
+				else
+					break;
 		}
 	}
 	else if (type == EV_ABS) {
@@ -1538,6 +1531,8 @@ static int toolTypeToDeviceType(WacomCommonPtr common, int type, int code)
 			case ABS_MT_SLOT:
 			case ABS_MT_TRACKING_ID:
 				return TOUCH_ID;
+			case ABS_MISC:
+				return usbFindDeviceTypeById(value);
 		}
 	}
 
@@ -1567,7 +1562,7 @@ static int refreshDeviceType(WacomCommonPtr common, int fd)
 	for (i = 0; i < KEY_MAX; i++)
 	{
 		if (ISBITSET(keys, i))
-			device_type = toolTypeToDeviceType(common, EV_KEY, i);
+			device_type = deviceTypeFromEvent(common, EV_KEY, i, 0);
 		if (device_type)
 			return device_type;
 	}
@@ -1575,6 +1570,35 @@ static int refreshDeviceType(WacomCommonPtr common, int fd)
 	return 0;
 }
 
+static int deriveDeviceTypeFromButtonEvent(WacomCommonPtr common,
+					   const struct input_event *event_ptr)
+{
+	wcmUSBData *usbdata = common->private;
+	int nkeys;
+
+	if (event_ptr->type == EV_KEY) {
+
+		switch (event_ptr->code) {
+		case BTN_LEFT:
+		case BTN_MIDDLE:
+		case BTN_RIGHT:
+		case BTN_SIDE:
+		case BTN_BACK:
+		case BTN_EXTRA:
+		case BTN_FORWARD:
+			return PAD_ID;
+		default:
+			for (nkeys = 0; nkeys < usbdata->npadkeys; nkeys++)
+			{
+				if (event_ptr->code == usbdata->padkey_code[nkeys]) {
+					return PAD_ID;
+				}
+			}
+		}
+	}
+	return 0;
+}
+
 /***
  * Retrieve the tool type from an USB data packet by looking at the event
  * codes. Refer to linux/input.h for event codes that define tool types.
@@ -1598,7 +1622,7 @@ static int usbInitToolType(WacomCommonPtr common, int fd,
 
 	for (i = 0; (i < nevents) && !device_type; ++i, event_ptr++)
 	{
-		device_type = toolTypeToDeviceType(common, event_ptr->type, event_ptr->code);
+		device_type = deviceTypeFromEvent(common, event_ptr->type, event_ptr->code, event_ptr->value);
 	}
 
 	if (!device_type)
@@ -1607,6 +1631,10 @@ static int usbInitToolType(WacomCommonPtr common, int fd,
 	if (!device_type)
 		device_type = refreshDeviceType(common, fd);
 
+	if (!device_type) /* expresskey pressed at startup or missing type */
+		for (i = 0; (i < nevents) && !device_type; ++i, event_ptr++)
+			device_type = deriveDeviceTypeFromButtonEvent(common, event_ptr);
+
 	return device_type;
 }
 
@@ -1634,8 +1662,8 @@ static void usbDispatchEvents(InputInfoPtr pInfo)
 	WacomDevicePtr priv = (WacomDevicePtr)pInfo->private;
 	WacomCommonPtr common = priv->common;
 	int channel;
-	WacomDeviceState dslast = common->wcmChannel[0].valid.state;
 	wcmUSBData* private = common->private;
+	WacomDeviceState dslast = common->wcmChannel[private->lastChannel].valid.state;
 
 	DBG(6, common, "%d events received\n", private->wcmEventCnt);
 
@@ -1670,23 +1698,15 @@ static void usbDispatchEvents(InputInfoPtr pInfo)
 		return;
 	}
 
-	/* Protocol 5 tools are dynamically assigned with channel numbers.
-	 * The structure must be initialized to known starting values
-	 * when first entering proximity to discard invalid data.
-	 */
-	if (common->wcmProtocolLevel == WCM_PROTOCOL_5)
-	{
-		if (!common->wcmChannel[channel].work.proximity)
-			memset(&common->wcmChannel[channel],0,
-			       sizeof(WacomChannel));
-	}
-
 	ds = &common->wcmChannel[channel].work;
 	dslast = common->wcmChannel[channel].valid.state;
 
+	if (ds->device_type && ds->device_type != private->wcmDeviceType)
+		LogMessageVerbSigSafe(X_ERROR, 0,
+				      "usbDispatchEvents: Device Type mismatch - %d -> %d. This is a BUG.\n",
+				      ds->device_type, private->wcmDeviceType);
 	/* no device type? */
 	if (!ds->device_type && private->wcmDeviceType) {
-		/* tool was on tablet at startup, force type and proximity */
 		ds->device_type = private->wcmDeviceType;
 		ds->proximity = 1;
 	}
@@ -1734,6 +1754,7 @@ static void usbDispatchEvents(InputInfoPtr pInfo)
 			if (event->code == REL_WHEEL)
 			{
 				ds->relwheel = -event->value;
+				ds->time = (int)GetTimeInMillis();
 				common->wcmChannel[channel].dirty |= TRUE;
 			}
 			else
@@ -1765,6 +1786,8 @@ static void usbDispatchEvents(InputInfoPtr pInfo)
 	if (!ds->proximity)
 		private->wcmLastToolSerial = 0;
 
+	private->lastChannel = channel;
+
 	for (c = 0; c < MAX_CHANNELS; c++) {
 		ds = &common->wcmChannel[c].work;
 
