@@ -89,23 +89,6 @@ static void wcmKbdCtrlCallback(DeviceIntPtr di, KeybdCtrl* ctrl)
 {
 }
 
-/*****************************************************************************
- * wcmVirtualTabletPadding(InputInfoPtr pInfo)
- ****************************************************************************/
-
-void wcmVirtualTabletPadding(InputInfoPtr pInfo)
-{
-	WacomDevicePtr priv = (WacomDevicePtr)pInfo->private;
-
-	priv->leftPadding = 0;
-	priv->topPadding = 0;
-
-	if (!is_absolute(pInfo)) return;
-
-	DBG(10, priv, "x=%d y=%d \n", priv->leftPadding, priv->topPadding);
-	return;
-}
-
 /*****************************************************************************
  * wcmInitialToolSize --
  *    Initialize logical size and resolution for individual tool.
@@ -128,12 +111,18 @@ wcmInitialToolSize(InputInfoPtr pInfo)
 	}
 	else
 	{
+		priv->minX = common->wcmMinX;
+		priv->minY = common->wcmMinY;
 		priv->maxX = common->wcmMaxX;
 		priv->maxY = common->wcmMaxY;
 		priv->resolX = common->wcmResolX;
 		priv->resolY = common->wcmResolY;
 	}
 
+	if (!priv->topX)
+		priv->topX = priv->minX;
+	if (!priv->topY)
+		priv->topY = priv->minY;
 	if (!priv->bottomX)
 		priv->bottomX = priv->maxX;
 	if (!priv->bottomY)
@@ -442,7 +431,6 @@ static int wcmDevInit(DeviceIntPtr pWcm)
 	if (!IsPad(priv))
 	{
 		wcmInitialToolSize(pInfo);
-		wcmMappingFactor(pInfo);
 	}
 
 	if (!wcmInitAxes(pWcm))
